from ckanext.versioned_datastore.logic.validators import not_empty, resource_id_exists


def vds_resource_check() -> dict:
    return {
        'resource_id': [not_empty, str, resource_id_exists],
    }



import bisect
from typing import List, Optional

from ckantools.decorators import action
from elasticsearch_dsl import Q
from splitgill.indexing.fields import DocumentField

from ckanext.versioned_datastore.lib.query.schema import get_schema_versions
from ckanext.versioned_datastore.lib.query.search.query import DirectQuery
from ckanext.versioned_datastore.lib.query.search.request import SearchRequest
from ckanext.versioned_datastore.lib.utils import get_database
from ckanext.versioned_datastore.logic.version import helptext, schema


@action(schema.vds_version_schema(), helptext.vds_version_schema, get=True)
def vds_version_schema():
    """
    Retrieves all the query schema versions that are available.

    :returns: a list of query schema versions.
    """
    return get_schema_versions()


@action(schema.vds_version_record(), helptext.vds_version_record, get=True)
def vds_version_record(resource_id: str, record_id: str) -> List[int]:
    """
    Retrieves all the versions available for a record in a resource. These will be
    returned as a list of ints in ascending order.

    :param resource_id: the resource ID
    :param record_id: the record ID
    :returns: a list of versions, or an empty list if the record doesn't exist
    """
    query = DirectQuery(
        [resource_id],
        None,
        Q('term', **{DocumentField.ID: record_id}),
    )
    request = SearchRequest(query, force_no_version=True)
    response = request.run()
    return sorted(hit.version for hit in response.hits)


@action(schema.vds_version_resource(), helptext.vds_version_resource, get=True)
def vds_version_resource(resource_id: str) -> List[int]:
    """
    Returns a list of the available versions for a datastore resource. These will be in
    ascending order.

    :param resource_id: the resource ID
    :returns: the versions in ascending order
    """
    database = get_database(resource_id)
    return database.get_versions()


@action(schema.vds_version_round(), helptext.vds_version_round, get=True)
def vds_version_round(resource_id: str, version: Optional[int] = None):
    """
    Given a resource ID and a version, rounds the version to the lowest, nearest version
    of the resource (i.e. rounds it down). If no version parameter is provided, or the
    provided version parameter higher than the current version of the resource, then the
    latest version of the resource is returned.

    :param resource_id: the resource ID
    :param version: the version to round (default is None)
    :returns: the rounded version
    """
    database = get_database(resource_id)
    versions = database.get_versions()

    if version is None or version >= versions[-1]:
        # cap the requested version to the latest version
        return versions[-1]
    elif version < versions[0]:
        # use the requested version if it's lower than the lowest available version
        return version
    else:
        # find the lowest, nearest version to the requested one
        position = bisect.bisect_right(versions, version)
        return versions[position - 1]



from typing import Optional

from ckan.plugins import toolkit
from ckantools.decorators import action
from ckantools.validators import list_of_dicts_validator
from elasticsearch_dsl import Q
from splitgill.indexing.fields import DocumentField

from ckanext.versioned_datastore.lib.importing.options import update_options
from ckanext.versioned_datastore.lib.importing.tasks import (
    queue_delete,
    queue_ingest,
    queue_sync,
)
from ckanext.versioned_datastore.lib.query.search.query import DirectQuery
from ckanext.versioned_datastore.lib.query.search.request import SearchRequest
from ckanext.versioned_datastore.lib.utils import (
    ReadOnlyResourceException,
    is_resource_read_only,
)
from ckanext.versioned_datastore.logic.data import helptext, schema


@action(schema.vds_data_add(), helptext.vds_data_add)
def vds_data_add(
    context: dict, original_data_dict: dict, resource_id: str, replace: bool
):
    """
    Add data to the datastore for the given resource. The data added will be the data
    found at the resource's URL unless a list of dict records is provided in which case
    this will be added. The addition is asynchronous and so the result returned by
    calling this action is information about the queued jobs. Two jobs will be queued,
    an ingest job to store the data found at the URL/in the provided records, and then a
    sync job, which updates the data search index making the data available for
    searching (this will only run if the ingest succeeded).

    If the replace flag is passed as True, all the existing data in the datastore for
    this resource will be deleted before the new data is added, resulting in the new
    data replacing the old data. This is the kind of behaviour users expect when
    uploading a fresh new file.

    Before the ingest and sync jobs are queued, the datastore's parsing options are
    updated if required. If they are altered before ingest, the new version will be
    returned in the result dict.

    :param context: the CKAN context
    :param resource_id: the resource ID
    :param replace: whether to replace all existing records with the new ones
    :returns: a dict containing information about the add
    """
    if is_resource_read_only(resource_id):
        raise ReadOnlyResourceException('This resource has been marked as read only')

    records = original_data_dict.get('records', None)
    if records is not None:
        records = list_of_dicts_validator(records, context)

    # make sure the options are in sync
    new_options_version = update_options(resource_id)

    user = toolkit.get_action('user_show')(context, {'id': context['user']})
    resource = toolkit.get_action('resource_show')(context, {'id': resource_id})
    ingest_job, sync_job = queue_ingest(resource, replace, user['apikey'], records)

    return {
        'new_options_version': new_options_version,
        'queued_at': ingest_job.enqueued_at.isoformat(),
        'ingest_job_id': ingest_job.id,
        'sync_job_id': sync_job.id,
    }


@action(schema.vds_data_delete(), helptext.vds_data_delete)
def vds_data_delete(context: dict, resource_id: str):
    """
    Deletes all the data from the given resource. This is an async operation as
    signified by the return of this action which is details about the created jobs. The
    data will be deleted by creating new deleted versions of the records in the resource
    and thus the old data will still be available in old versions, but the latest
    version of the resource will be empty.

    Two jobs will be queued, one to delete all the records in MongoDB and then one to
    sync the changes with Elasticsearch.

    :param context: the CKAN action context
    :param resource_id: the resource ID to delete the data from
    :returns: a dict containing information about the delete
    """
    if is_resource_read_only(resource_id):
        raise ReadOnlyResourceException('This resource has been marked as read only')

    resource = toolkit.get_action('resource_show')(context, {'id': resource_id})
    delete_job, sync_job = queue_delete(resource)

    return {
        'queued_at': delete_job.enqueued_at.isoformat(),
        'delete_job_id': delete_job.id,
        'sync_job_id': sync_job.id,
    }


@action(schema.vds_data_sync(), helptext.vds_data_sync)
def vds_data_sync(context: dict, resource_id: str, full: bool = False):
    """
    Queues as sync operation to ensure that the resource's data in MongoDB matches the
    data in Elasticsearch.

    With full=False (the default), only any unsynced changes will be synced. With
    full=True, all records will be resynced. This is managed by Splitgill and may delete
    records before syncing (check Splitgill docs).

    :param context: the CKAN action context
    :param resource_id: the resource ID to sync
    :param full: whether to sync all records again or just the changed ones since the
        last sync (default is False).
    :returns: a dict containing information about the sync
    """
    if is_resource_read_only(resource_id):
        raise ReadOnlyResourceException('This resource has been marked as read only')

    resource = toolkit.get_action('resource_show')(context, {'id': resource_id})
    job = queue_sync(resource, full)

    return {
        'queued_at': job.enqueued_at.isoformat(),
        'job_id': job.id,
    }


# for compat reasons
@action(schema.vds_data_get(), helptext.vds_data_get, get=True)
def record_show(resource_id: str, record_id: str, version: Optional[int] = None):
    """
    Compatibility action for vds_data_get.

    :param resource_id: the resource's ID
    :param record_id: the record's ID
    :param version: the version to get of the record, or None to get the latest
    :returns: the record data
    """
    return vds_data_get(resource_id, record_id, version)


@action(schema.vds_data_get(), helptext.vds_data_get, get=True)
def vds_data_get(resource_id: str, record_id: str, version: Optional[int] = None):
    """
    Retrieves the data for one record from the given resource at the given optional
    version and returns it. If the record can't be found an ObjectNotFound exception is
    raised.

    :param resource_id: the resource's ID
    :param record_id: the record's ID
    :param version: the version to get of the record, or None to get the latest
    :returns:
    """
    query = DirectQuery(
        [resource_id], version, Q('term', **{DocumentField.ID: record_id})
    )
    request = SearchRequest(query, size=1)
    response = request.run()
    try:
        return {
            'data': response.data[0],
            # TODO: do we actually need to add the fields?
            # get_fields(resource_id, request.query.version)
            'resource_id': resource_id,
        }
    except IndexError:
        # if we don't have a result, raise not found
        raise toolkit.ObjectNotFound



from ckan.plugins import toolkit
from ckantools.decorators import action

from ckanext.versioned_datastore.lib.downloads.download import DownloadRunManager
from ckanext.versioned_datastore.lib.downloads.notifiers import validate_notifier_args
from ckanext.versioned_datastore.logic.download import helptext, schema
from ckanext.versioned_datastore.logic.download.arg_objects import (
    DerivativeArgs,
    NotifierArgs,
    QueryArgs,
    ServerArgs,
)
from ckanext.versioned_datastore.model.downloads import DownloadRequest


@action(schema.vds_download_queue(), helptext.vds_download_queue)
def vds_download_queue(
    context,
    query: QueryArgs,
    file: DerivativeArgs,
    server: ServerArgs = None,
    notifier: NotifierArgs = None,
):
    """
    Queues a download and returns information about the job and status.

    :param context: the CKAN action context
    :param query: the query as a QueryArgs object
    :param file: the file options as a DerivativeArgs object
    :param server: the server options as a ServerArgs object (default is None)
    :param notifier: the notifier options as a NotifierArgs object (default is None)
    :returns: a dict of information about the download
    """
    server = server or ServerArgs(**ServerArgs.defaults)
    notifier = notifier or NotifierArgs(**NotifierArgs.defaults)

    validate_notifier_args(notifier.type, notifier.type_args)

    if server.custom_filename:
        # check if admin
        try:
            toolkit.check_access('vds_custom_download_filename', context)
        except toolkit.NotAuthorized:
            server.custom_filename = None

    download_runner = DownloadRunManager(
        query_args=query,
        derivative_args=file,
        server_args=server,
        notifier_args=notifier,
    )

    job = toolkit.enqueue_job(
        download_runner.run,
        queue='download',
        title=download_runner.request.created.strftime('%Y-%m-%d %H:%M:%S'),
        rq_kwargs={'timeout': '24h'},
    )

    return {
        'queued_at': job.enqueued_at.isoformat(),
        'job_id': job.id,
        'download_id': download_runner.request.id,
        'status_json': toolkit.url_for(
            'datastore_status.download_status_json',
            download_id=download_runner.request.id,
            qualified=True,
        ),
        'status_html': toolkit.url_for(
            'datastore_status.download_status',
            download_id=download_runner.request.id,
            qualified=True,
        ),
    }


@action(schema.vds_download_regenerate(), helptext.vds_download_regenerate)
def vds_download_regenerate(
    context,
    download_id,
    server: ServerArgs = None,
    notifier: NotifierArgs = None,
):
    """
    Regenerates a download.

    :param context: the CKAN action context
    :param download_id: the download ID
    :param server: the server options as a ServerArgs object (default is None)
    :param notifier: the notifier options as a NotifierArgs object (default is None)
    :returns: a dict of information about the download
    """
    # find the download request first
    original_request = DownloadRequest.get(download_id)
    if original_request is None:
        raise toolkit.ObjectNotFound(f'Download "{download_id}" cannot be found.')

    server_args = ServerArgs.defaults.copy()
    if original_request.server_args is not None:
        server_args.update(**original_request.server_args)
    server = server or ServerArgs(**server_args)

    notifier = notifier or NotifierArgs(**NotifierArgs.defaults)

    query = QueryArgs(
        query=original_request.core_record.query,
        query_version=original_request.core_record.query_version,
        resource_ids=original_request.core_record.get_resource_ids(),
        version=original_request.core_record.get_version(),
    )

    file = DerivativeArgs(
        format=original_request.derivative_record.format,
        **original_request.derivative_record.options,
    )

    return vds_download_queue(context, query, file, server, notifier)



from collections import defaultdict
from typing import List, Optional

from ckantools.decorators import action
from elasticsearch.exceptions import NotFoundError
from elasticsearch_dsl import A, MultiSearch, Q, Search
from splitgill.search import keyword, number, version_query

from ckanext.versioned_datastore.lib.query.schema import (
    get_latest_query_version,
    hash_query,
    validate_query,
)
from ckanext.versioned_datastore.lib.utils import (
    es_client,
    get_database,
    ivds_implementations,
    unprefix_index_name,
)
from ckanext.versioned_datastore.logic.multi import helptext, schema
from ckanext.versioned_datastore.logic.multi.groups import FieldGroups
from ckanext.versioned_datastore.logic.multi.utils import (
    get_available_datastore_resources,
    make_request,
)


@action(schema.vds_multi_query(), helptext.vds_multi_query, get=True)
def vds_multi_query(data_dict: dict):
    """
    Queries the given resources using the given query and at an optional version, and
    then return the results. Compared to the basic query, this action provides a richer
    query language and the ability to query multiple resources at the same time.

    :param data_dict: the data dict of options
    :returns: a dict which contains the total number of records found, an after value
        for pagination, and a list of dicts of record data
    """
    request = make_request(data_dict)
    response = request.run()
    result = {
        'total': response.count,
        'after': response.next_after,
        'records': [
            {
                'data': hit.data,
                'resource': hit.resource_id,
            }
            for hit in response.hits
        ],
    }

    for plugin in ivds_implementations():
        plugin.vds_after_multi_query(response, result)

    return result


@action(schema.vds_multi_count(), helptext.vds_multi_count, get=True)
def vds_multi_count(data_dict: dict):
    """
    Queries the given resources using the given query and at an optional version, and
    then returns the total number of records which matched the query and the counts per
    resource. Compared to the basic query count, this action provides a richer query
    language and the ability to query multiple resources at the same time.

    :param data_dict: the data dict of options
    :returns: a dict which contains the total count and a breakdown of the hits per
        resource
    """
    request = make_request(data_dict)
    request.set_no_results()
    # use an aggregation to get the hit count of each resource, set the size to the
    # number of resources we're querying to ensure we get all counts in one go and don't
    # have to paginate with a composite agg
    request.add_agg(
        'counts', 'terms', field='_index', size=len(request.query.resource_ids)
    )
    response = request.run()

    # default the counts to 0 for all resources
    counts = {resource_id: 0 for resource_id in request.query.resource_ids}
    # then update with the counts from the resources that matched the query
    counts.update(
        {
            unprefix_index_name(bucket['key']): bucket['doc_count']
            for bucket in response.aggs['counts']['buckets']
        }
    )
    return {'total': response.count, 'counts': counts}


@action(
    schema.vds_multi_autocomplete_value(),
    helptext.vds_multi_autocomplete_value,
    get=True,
)
def vds_multi_autocomplete_value(
    data_dict: dict,
    field: str,
    prefix: Optional[str] = None,
):
    """
    Queries the given resources using the given query and at an optional version, and
    then returns the values that match the given prefix for the given field. The list of
    values is returned.

    :param data_dict: the action options
    :param field: the field to get the values from
    :param prefix: the prefix to search for
    :returns: a dict containing the list of values and an after key if there are more
        values available. If the after key is not present, there are no more values
        available
    """
    # extract the limit but default to a size of 20 if it's not present
    size = data_dict.pop('size', 20)
    # grab the after as we need to use it for the agg, not the query
    after = data_dict.pop('after', None)

    request = make_request(data_dict)
    request.set_no_results()

    # create the full path to the parsed field type we're going to filter and agg over
    field_path = keyword(field)

    if prefix:
        request.extra_filter &= Q('prefix', **{field_path: prefix})

    # add the aggregation which gets the field values
    request.add_agg(
        'field_values',
        'composite',
        # get one more than the requested size so that we can work out the after
        size=size + 1,
        sources={field: A('terms', field=field_path, order='asc')},
        # only include the after key if there is one
        **({'after': {field: after}} if after is not None else {}),
    )

    response = request.run()

    agg_result = response.aggs['field_values']
    values = [bucket['key'][field] for bucket in agg_result['buckets']]
    result = {'values': values[:size]}
    if 'after_key' in agg_result and len(values) > size:
        result['after'] = agg_result['after_key'][field]
    return result


@action(
    schema.vds_multi_autocomplete_field(),
    helptext.vds_multi_autocomplete_field,
    get=True,
)
def vds_multi_autocomplete_field(
    resource_ids: List[str],
    text: str = '',
    lowercase: bool = False,
    version: Optional[int] = None,
):
    """
    Queries the field names in the given resources, filtering by checking if the field
    name contains the text parameter.

    The fields key in the response dict contains the field's full path, and then for
    each resource ID the field path appears in:
        - the field's name
        - the field's path
        - the number of records which have this field
        - the number of records which have a text parsed value in this field
        - the number of records which have a keyword parsed value in this field
        - the number of records which have a boolean parsed value in this field
        - the number of records which have a date parsed value in this field
        - the number of records which have a number parsed value in this field
        - the number of records which have a geo parsed value in this field

    Note that for large numbers of resources, this action is quite expensive to run.

    :param resource_ids: the resources match fields on (if no resource IDs are passed,
                         all resources are searched)
    :param text: the text to search for
    :param lowercase: whether to compare the text to the field names in lowercase
                      (default is False)
    :param version: the version to search at (default is None, which means latest)
    :returns: the total number of fields matched and details about the fields that were
             matched
    """
    fields = defaultdict(dict)

    # if no resource IDs were provided, use all resources available to the user
    if not resource_ids:
        resource_ids = sorted(get_available_datastore_resources())

    for resource_id in resource_ids:
        database = get_database(resource_id)

        try:
            parsed_fields = database.get_parsed_fields(version=version)
        except NotFoundError:
            # temporary fix for splitgill#38 (so we can ignore unavailable resources)
            continue

        for field in parsed_fields:
            if text in (field.path.lower() if lowercase else field.path):
                fields[field.path][resource_id] = {
                    'name': field.name,
                    'path': field.path,
                    'count': field.count,
                    'text': field.count_text,
                    'keyword': field.count_keyword,
                    'boolean': field.count_boolean,
                    'date': field.count_date,
                    'number': field.count_number,
                    'geo': field.count_geo,
                }

    return {'count': len(fields), 'fields': fields}


@action(schema.vds_multi_hash(), helptext.vds_multi_hash, get=True)
def vds_multi_hash(query: dict, query_version: Optional[str] = None):
    """
    Given a query and a query schema version, hash them and return it.

    :param query: the query
    :param query_version: the version of the query schema
    :returns: the hash
    """
    if query_version is None:
        query_version = get_latest_query_version()
    validate_query(query, query_version)
    return hash_query(query, query_version)


@action(schema.vds_multi_fields(), helptext.vds_multi_fields, get=True)
def vds_multi_fields(
    data_dict: dict, size: int = 10, ignore_groups: Optional[List[str]] = None
):
    """
    Groups the fields available on the given resources at the optional given version and
    returns as many of them as requested in the size parameter. The groups are created
    by matching field names across resources and then sorted in descending order with
    the most common groups at the top (both most common in terms of resources containing
    the field, but also most common as in with the most records that have the field in
    them in a resource).

    :param data_dict: the action options
    :param size: the number of field groups to return (defaults to the top 10)
    :param ignore_groups: an optional list of fields to ignore
    :returns: a list of field groups represented as dicts
    """
    request = make_request(data_dict)
    request.set_no_results()

    query = request.query.to_dsl()

    field_groups = FieldGroups(skip_ids=True)
    if ignore_groups:
        for ignore in ignore_groups:
            field_groups.ignore(ignore)

    for plugin in ivds_implementations():
        plugin.vds_modify_field_groups(request.query.resource_ids, field_groups)

    for resource_id in request.query.resource_ids:
        database = get_database(resource_id)
        try:
            fields = database.get_parsed_fields(
                version=request.query.version, query=query
            )
        except NotFoundError:
            # temporary fix for splitgill#38 (so we can ignore unavailable resources)
            continue
        field_groups.add(resource_id, fields)

    return field_groups.select(size)


@action(schema.vds_multi_stats(), helptext.vds_multi_stats, get=True)
def vds_multi_stats(data_dict: dict, field: str, missing: Optional[float] = None):
    """
    Retrieves a simple set of numerical stats about the given field and returns them in
    a dict. The stats provided are the same as the Elasticsearch stats aggregation
    (because that is what is used!), so you'll get the following keys in the dict
    response:

      - count
      - min
      - max
      - avg
      - sum

    The missing parameter defines how documents that are missing a value should be
    treated. By default, they will be ignored, but it is also possible to treat them as
    if they had a value by providing one here.

    :param data_dict: the data dict passed to this action
    :param field: the field to get stats for
    :param missing: value to use for records missing this field, or None to ignore them
    :returns: a dict of statistical data
    """
    request = make_request(data_dict)
    request.set_no_results()
    agg_options = {'field': number(field)}
    if missing is not None:
        agg_options['missing'] = missing
    request.add_agg('field_stats', 'stats', **agg_options)
    response = request.run()
    return response.aggs['field_stats']


@action(schema.vds_multi_direct(), helptext.vds_multi_direct)
def vds_multi_direct(data_dict: dict):
    """
    Allows users to run Elasticsearch queries directly against the cluster. The raw
    response from Elasticsearch is returned directly as well. This is locked down to
    admin users only to avoid misuse, but can be used from other extensions easily with
    the ignore_auth context option.

    To control the version that is searched, either include it in the search object
    directly, or:
        - pass "latest", None, or don't include the version key in the data dict to
          search the latest version
        - pass "all" to search all versions
        - pass a version timestamp value to search at a specific version

    A list of resource IDs must be passed and if no resource IDs are passed an error is
    raised during validation.

    :param data_dict: the action data dict
    :returns: the raw response from Elasticsearch
    """
    resource_ids = data_dict['resource_ids']
    version = data_dict.get('version', 'latest')
    search = Search.from_dict(data_dict.get('search', {}))

    databases = map(get_database, resource_ids)
    if version == 'latest':
        indices = [database.indices.latest for database in databases]
    else:
        indices = [database.indices.wildcard for database in databases]
        if version is not None and version != 'all':
            search = search.filter(version_query(int(version)))

    # call search.index() empty first to reset the indices thus avoiding them being set
    # in the passed search dict
    search = search.index().index(indices)

    multi_search = MultiSearch(using=es_client()).add(search)
    result = next(iter(multi_search.execute()))

    return result.to_dict()



from typing import Optional

from ckan.plugins import toolkit
from ckantools.decorators import action
from elasticsearch_dsl import A, Q
from splitgill.search import ALL_POINTS, keyword

from ckanext.versioned_datastore.logic.basic import helptext, schema
from ckanext.versioned_datastore.logic.basic.utils import (
    format_facets,
    get_fields,
    make_request,
)


# for compat reasons
@action(schema.vds_basic_query(), helptext.vds_basic_query, get=True)
def datastore_search(
    data_dict: dict,
    resource_id: str,
    run_query: Optional[bool] = True,
):
    """
    Performs a basic query on one resource with parameters and return value compatible
    with CKAN's datastore.

    Just calls vds_basic_query.
    """
    return vds_basic_query(data_dict, resource_id, run_query)


@action(schema.vds_basic_query(), helptext.vds_basic_query, get=True)
def vds_basic_query(
    data_dict: dict,
    resource_id: str,
    run_query: Optional[bool] = True,
):
    """
    Performs a basic query on one resource using a simple query language which is in
    line with CKAN's default datastore extension.

    :param data_dict: the data dict passed to this action
    :param resource_id: the ID of the resource to search
    :param run_query: whether to run the query, or just return information about what
        would have been run
    :returns: a dict
    """
    request = make_request(data_dict)
    if not run_query:
        return {'indexes': request.indexes(), 'search': request.to_search().to_dict()}

    response = request.run()
    return {
        'total': response.count,
        'records': response.data,
        'facets': format_facets(response.aggs),
        'fields': get_fields(resource_id, request.query.version),
        'after': response.next_after,
    }


@action(schema.vds_basic_count(), helptext.vds_basic_query, get=True)
def vds_basic_count(data_dict: dict):
    """
    Counts the number of records that meet the given query on the resource with the
    given ID and returns this number.

    :param data_dict: the data dict passed to this action
    :returns: an integer >= 0
    """
    request = make_request(data_dict)
    request.set_no_results()
    response = request.run()
    return response.count


@action(schema.vds_basic_autocomplete(), helptext.vds_basic_autocomplete, get=True)
def vds_basic_autocomplete(data_dict: dict, field: str, term: str):
    """
    Provides autocomplete for values on the given field in the given resource under the
    given query and version restrictions. Use after and limit to paginate through
    results.

    :param data_dict: the data dict passed to this action
    :param field: the field to autocomplete values on
    :param term: a prefix to filter field values by
    :returns: a dict containing the values and optionally an after key for pagination
    """
    # extract the limit but default to a size of 20 if it's not present
    size = data_dict.pop('limit', 20)
    if not 0 < size <= 20:
        raise toolkit.ValidationError('Size must be 0 < size <= 20')
    # grab the after as we need to use it for the agg, not the query
    after = data_dict.pop('after', None)

    request = make_request(data_dict)
    request.set_no_results()
    # add the autocomplete part of the query
    request.extra_filter &= Q('prefix', **{keyword(field): term})

    # add the aggregation which gets the field values
    agg_options = dict(field=keyword(field), order='asc')
    if after is not None:
        agg_options['after'] = {field: after}
    request.add_agg(
        'field_values',
        'composite',
        size=size,
        sources={field: A('terms', **agg_options)},
    )

    response = request.run()
    agg_result = response.aggs['field_values']
    return_dict = {
        'values': [bucket['key'][field] for bucket in agg_result['buckets']],
    }
    if 'after_key' in agg_result:
        return_dict['after'] = agg_result['after_key'][field]
    return return_dict


# for compat reasons
@action(schema.vds_basic_query(), helptext.vds_basic_extent, get=True)
def datastore_query_extent(data_dict: dict):
    """
    Performs a basic query extent analysis on one resource with parameters and return
    value compatible with CKAN's datastore.

    Just calls vds_basic_extent.
    """
    return vds_basic_extent(data_dict)


@action(schema.vds_basic_extent(), helptext.vds_basic_extent, get=True)
def vds_basic_extent(data_dict: dict):
    """
    Calculates the geographic extent of the given resource using the given query.
    Returns a dict containing the overall number of records which matched the query, the
    number of records with geo data that matched the query and then the bounds as a pair
    of coordinates (top left, bottom right).

    :param data_dict: the data dict passed to this action
    :returns: a dict
    """
    request = make_request(data_dict)
    request.set_no_results()

    # get the total number of records first before we do anything else
    total = request.run().count

    # add a filter to get only records with geo data
    request.extra_filter &= Q('exists', field=ALL_POINTS)
    # add our bounds aggregation
    request.add_agg('bounds', 'geo_bounds', field=ALL_POINTS, wrap_longitude=False)

    response = request.run()

    result = {
        'total_count': total,
        'geom_count': response.count,
    }
    # extract and add the bounds info from the aggregations if there is any
    if response.count > 0:
        top_left = response.aggs['bounds']['bounds']['top_left']
        bottom_right = response.aggs['bounds']['bounds']['bottom_right']
        result['bounds'] = [[p['lat'], p['lon']] for p in (top_left, bottom_right)]

    return result



import abc
import json
from operator import itemgetter
from typing import Any, List, Optional, Union

from ckan.plugins import toolkit
from elasticsearch_dsl import Q
from elasticsearch_dsl.query import EMPTY_QUERY, Bool
from elasticsearch_dsl.query import Query as DSLQuery
from splitgill.search import ALL_POINTS, match_query, term_query

from ckanext.versioned_datastore.lib.query.schema import (
    get_latest_query_version,
    hash_query,
    normalise_query,
    translate_query,
    validate_query,
)


class Query(abc.ABC):
    """
    Abstract base class representing a query at a specific version over a specific set
    of resources.

    Subclasses should override the to_dsl method to provide customised filtering
    functionality.
    """

    def __init__(
        self, kind: str, resource_ids: List[str], version: Optional[int] = None
    ):
        """
        :param kind: an identifier for the kind of query this is
        :param resource_ids: the IDs of the resources to search
        :param version: the version at which to search
        """
        self.kind = kind
        self.resource_ids = resource_ids
        self.version = version

    @abc.abstractmethod
    def to_dsl(self) -> DSLQuery:
        """
        Returns just an Elasticsearch DSL Query object representing this search.

        :returns: an Elasticsearch Query instance
        """
        ...

    def __eq__(self, other: Any):
        if isinstance(other, Query):
            return self.to_dsl() == other.to_dsl()
        return False


class DirectQuery(Query):
    """
    A query created using an Elasticsearch DSL Query object.
    """

    def __init__(
        self,
        resource_ids: List[str],
        version: Optional[int] = None,
        dsl_query: Optional[DSLQuery] = None,
    ):
        """
        :param resource_ids: the IDs of the resources to search
        :param version: the version to search at
        :param dsl_query: the query to perform
        """
        super().__init__('direct', resource_ids, version)
        self.dsl = dsl_query if dsl_query is not None else EMPTY_QUERY

    def to_dsl(self) -> DSLQuery:
        return self.dsl


class SchemaQuery(Query):
    """
    A query using the schemas registered with this plugin.
    """

    def __init__(
        self,
        resource_ids: List[str],
        version: Optional[int] = None,
        query: Optional[dict] = None,
        query_version: Optional[str] = None,
    ):
        """
        :param resource_ids: the IDs of the resources to search
        :param version: the version to search at
        :param query: the query to use (defaults to {} if not given)
        :param query_version: the query version to use (defaults to latest if not given)
        """
        super().__init__('schema', resource_ids, version)
        self.query_version = (
            query_version if query_version else get_latest_query_version()
        )
        if query:
            self.query = normalise_query(query, self.query_version)
        else:
            self.query = {}

    @property
    def hash(self) -> str:
        """
        Returns the hash of the query for use in identifying it uniquely.

        :returns: the hash as a str
        """
        return hash_query(self.query, self.query_version)

    def validate(self):
        """
        Checks that the query is valid according to the schema.

        If valid, nothing happens, if not valid, an error will be raised.
        """
        validate_query(self.query, self.query_version)

    def to_dsl(self) -> DSLQuery:
        """
        Checks that the query is valid according to the schema and then translates it to
        a Search object.

        :returns: a Search object
        """
        self.validate()
        return translate_query(self.query, self.query_version)


class BasicQuery(Query):
    """
    A query using the old CKAN datastore style query method (with a couple of minor
    differences).
    """

    def __init__(
        self,
        resource_id: str,
        version: Optional[int] = None,
        q: Optional[Union[str, dict]] = None,
        filters: Optional[dict] = None,
    ):
        """
        :param resource_id: the ID of the resource to search
        :param version: the version to search at
        :param q: a query string which will be searched against the all text field or a
                  dict of fields and search values. If this is a dict then the keys
                  are understood as field names, but if a key is an empty string then
                  the all text field is searched. This allows combination searches
                  across all text and specific fields.
        :param filters: a dict of fields and values to filter the result with. If a key
                        is present that is equal to "__geo__" then the value associated
                        with it should be a dict which will be treated as a geo query to
                        be run against the all points field. The value should contain a
                        "type" key which must have a corresponding value of "point",
                        "box" or "polygon" and then other keys that are dependent on the
                        type:
                          - point:
                            - distance: the radius of the circle centred on the
                                        specified location within which records must lie
                                        to be matched. This can specified in any form
                                        that Elasticsearch accepts for distances (see
                                        their doc, but values like 10km etc).
                            - point: the point to centre the radius on, specified as a
                                     lat, long pair in a list (i.e. [-20, 40.2]).
                        - box:
                          - points: the top left and bottom right points of the box,
                                    specified as a list of two lat/long pairs
                                    (i.e. [[-20, 40.2], [0.5, 100]]).
                        - polygon:
                          - points: a list of at least 3 lat/long pairs
                                    (i.e. [[-1, 4], [-1.1, 4.8], [1.99, 35], [5, 49]]).
        """
        super().__init__('basic', [resource_id], version)
        self.q = q
        self.filters = filters

    def to_dsl(self) -> DSLQuery:
        """
        Converts the q and filters parameters specified on creation to an Elasticsearch
        DSL query.

        :returns: the Elasticsearch DSL query
        """
        must = []
        if self.q is not None and self.q != '' and self.q != {}:
            if isinstance(self.q, (str, int, float)):
                # add a free text query across all fields
                must.append(match_query(str(self.q), operator='and'))
            elif isinstance(self.q, dict):
                for field, query in sorted(self.q.items(), key=itemgetter(0)):
                    must.append(
                        match_query(
                            str(query),
                            # todo: use __all__ instead of empty str to match __geo__?
                            field=None if field == '' else field,
                            operator='and',
                        )
                    )

        filters = []
        if self.filters is not None:
            for field, values in sorted(self.filters.items(), key=itemgetter(0)):
                if not isinstance(values, list):
                    values = [values]
                if field == '__geo__':
                    # only pass through the first value
                    filters.append(BasicQuery.create_geo_filter(values[0]))
                else:
                    for value in values:
                        filters.append(term_query(field, str(value)))

        # return the cleanest version of the search we can
        if not must and not filters:
            return EMPTY_QUERY
        elif must and not filters:
            return must[0] if len(must) == 1 else Bool(must=must)
        elif not must and filters:
            return filters[0] if len(filters) == 1 else Bool(filter=filters)
        else:
            return Bool(must=must, filter=filters)

    @staticmethod
    def create_point_filter(distance: str, coordinates: list) -> DSLQuery:
        """
        Creates an Elasticsearch Query object filtering for records within the radius of
        the given point coordinates.

        :param distance: the radius of the circle centred on the specified location
            within which records must lie to be matched. This can be specified in any
            form that elasticsearch accepts for distances (see their doc, but
            essentially values like 10km etc).
        :param coordinates: the point to centre the radius on, specified as a lon/lat
            pair in a list (i.e. [40.2, -20]).
        :returns: a Query object
        """
        options = {
            'distance': distance,
            ALL_POINTS: {
                'lat': float(coordinates[1]),
                'lon': float(coordinates[0]),
            },
        }
        return Q('geo_distance', **options)

    @staticmethod
    def create_multipolygon_filter(coordinates: list) -> DSLQuery:
        """
        Creates a Query object for the given multipolygon filters. Only the first group
        in each polygon grouping will be used as elasticsearch doesn't support this type
        of query yet (this kind of query is used for vacating space inside a polygon,
        like a donut for example).

        If more than one group is included then they are included as an or with a
        minimum must match of 1.

        :param coordinates: a list of a list of a list of a list of at least 3 lon/lat
            pairs (i.e. [[[[-16, 44], [-13.1, 34.8], [15.99, 35], [5, 49]]]])
        :returns: a Query object
        """
        should = []
        for group in coordinates:
            points = group[0]
            if len(points) < 3:
                raise toolkit.ValidationError(
                    'Not enough points in the polygon, must be 3 or more'
                )

            options = {
                ALL_POINTS: {
                    'shape': {
                        'type': 'Polygon',
                        # format the polygon point data appropriately
                        'coordinates': [
                            [[float(point[0]), float(point[1])] for point in points]
                        ],
                    }
                },
            }
            should.append(Q('geo_shape', **options))

        if len(should) == 1:
            # no point in producing an or query if we don't need to
            return should[0]
        else:
            # otherwise, create an or over the series of polygons
            return Bool(should=should, minimum_should_match=1)

    @staticmethod
    def create_polygon_filter(coordinates: list) -> DSLQuery:
        """
        Creates a polygon Query object. Only the first group in each polygon grouping
        will be used as elasticsearch doesn't support this type of query yet (this kind
        of query is used for vacating space inside a polygon, like a donut for example.

        If more than one group is included then they are included as an or with a
        minimum must match of 1.

        :param coordinates: a list of a list of a list of at least 3 lon/lat pairs (i.e.
            [[[-16, 44], [-13.1, 34.8], [15.99, 35], [5, 49]]])
        :returns: a Query object
        """
        # just wrap in another list and pass to the multipolygon handler
        return BasicQuery.create_multipolygon_filter([coordinates])

    @staticmethod
    def create_geo_filter(geo_filter: Union[str, dict]) -> DSLQuery:
        """
        Creates a Query object for the geo filter specified in the geo_filter dict.

        :param geo_filter: a dict describing a geographic filter. This should be a
            GeoJSON geometry and should therefore include a type key and a coordinates
            key. The type must be one of: Point, MultiPolygon or Polygon. In the case of
            a Point, a distance key is also required which specifies the radius of the
            point in a form elasticsearch understands (for example, 10km).
        :returns: a Query object
        """
        # we support 3 GeoJSON types currently, Point, MultiPolygon and Polygon
        query_type_map = {
            'Point': (BasicQuery.create_point_filter, {'distance', 'coordinates'}),
            'MultiPolygon': (BasicQuery.create_multipolygon_filter, {'coordinates'}),
            'Polygon': (BasicQuery.create_polygon_filter, {'coordinates'}),
        }

        try:
            # if it hasn't been parsed, parse the geo_filter as JSON
            if not isinstance(geo_filter, dict):
                geo_filter = json.loads(geo_filter)
            # fetch the function which will build the query into the search object
            create_function, required_params = query_type_map[geo_filter['type']]
        except (TypeError, ValueError):
            raise toolkit.ValidationError(
                'Invalid geo filter information, must be JSON'
            )
        except KeyError:
            raise toolkit.ValidationError(
                'Invalid query type, must be point, box or polygon'
            )

        try:
            # try and pull out the required parameters for each type
            parameters = {param: geo_filter[param] for param in required_params}
        except KeyError:
            required = ', '.join(required_params)
            raise toolkit.ValidationError(
                f'Missing parameters, must include {required}'
            )

        # call the function which will update the search object to include the geo
        # filters and then return the resulting search object
        return create_function(**parameters)



from collections import Counter

import pytest
from splitgill.indexing.fields import DataField, ParsedField, ParsedType
from splitgill.model import Record

from ckanext.versioned_datastore.lib.importing.options import (
    create_default_options_builder,
)
from ckanext.versioned_datastore.lib.query.search.query import BasicQuery
from ckanext.versioned_datastore.lib.query.search.sort import Sort
from ckanext.versioned_datastore.lib.utils import get_database
from ckanext.versioned_datastore.logic.basic.utils import (
    find_version,
    format_facets,
    get_fields,
    infer_type,
    make_request,
)


def test_format_facets():
    # first, check it can deal with an empty aggregation result
    assert format_facets({}) == {}

    facets = format_facets(
        {
            'facet1': {
                'sum_other_doc_count': 901,
                'doc_count_error_upper_bound': 12,
                'buckets': [
                    {'key': 'value1', 'doc_count': 43},
                    {'key': 'value2', 'doc_count': 243},
                    {'key': 'value3', 'doc_count': 543},
                    {'key': 'value4', 'doc_count': 143},
                    {'key': 'value5', 'doc_count': 743},
                ],
            },
            'facet2': {
                'sum_other_doc_count': 0,
                'doc_count_error_upper_bound': 0,
                'buckets': [
                    {'key': 'value1', 'doc_count': 6},
                ],
            },
        }
    )

    assert len(facets) == 2
    assert facets['facet1']['details']['sum_other_doc_count'] == 901
    assert facets['facet1']['details']['doc_count_error_upper_bound'] == 12
    assert len(facets['facet1']['values']) == 5
    assert facets['facet1']['values']['value1'] == 43
    assert facets['facet1']['values']['value2'] == 243
    assert facets['facet1']['values']['value3'] == 543
    assert facets['facet1']['values']['value4'] == 143
    assert facets['facet1']['values']['value5'] == 743

    assert facets['facet2']['details']['sum_other_doc_count'] == 0
    assert facets['facet2']['details']['doc_count_error_upper_bound'] == 0
    assert len(facets['facet2']['values']) == 1
    assert facets['facet2']['values']['value1'] == 6


@pytest.mark.filterwarnings('ignore::sqlalchemy.exc.SADeprecationWarning')
@pytest.mark.usefixtures('with_vds')
def test_get_fields():
    database = get_database('test')
    database.update_options(create_default_options_builder().build(), commit=False)
    database.ingest(
        [Record.new({'field1': 'egg', 'field2': 52, 'field3': [1, 2, 3]})], commit=True
    )
    database.sync()

    fields = get_fields('test')
    assert fields == [
        {'id': '_id', 'type': 'string'},
        {'id': 'field1', 'type': 'string', 'sortable': True},
        {'id': 'field2', 'type': 'number', 'sortable': True},
        {'id': 'field3', 'type': 'array', 'sortable': False},
    ]


class TestInferType:
    def test_non_parsed(self):
        df = DataField('test')
        pfs = {}
        assert infer_type(df, pfs) == 'object'

    def test_threshold(self):
        df = DataField('test')
        pf = ParsedField('test', count=10, type_counts=Counter({ParsedType.NUMBER: 5}))
        pfs = {'test': pf}
        assert infer_type(df, pfs, threshold=0.9) == 'string'
        assert infer_type(df, pfs, threshold=0.3) == 'number'
        assert infer_type(df, pfs, threshold=0.5) == 'number'

    def test_order_of_preference(self):
        df = DataField('test')
        pf = ParsedField('test', count=10, type_counts=Counter())
        pfs = {'test': pf}

        assert infer_type(df, pfs) == 'string'

        pf.type_counts[ParsedType.NUMBER] = 10
        assert infer_type(df, pfs) == 'number'

        pf.type_counts[ParsedType.BOOLEAN] = 10
        assert infer_type(df, pfs) == 'boolean'

        pf.type_counts[ParsedType.DATE] = 10
        assert infer_type(df, pfs) == 'date'


class TestFindVersion:
    plugins = ['versioned_datastore']

    def test_none_found(self):
        assert find_version({}) is None

    def test_version_is_none(self):
        assert find_version({'version': None}) is None

    def test_version_exists(self):
        assert find_version({'version': 10}) == 10
        assert find_version({'version': '10'}) == 10

        with pytest.raises(ValueError):
            find_version({'version': 'aaaa'})

    def test_version_as_filter_is_none(self):
        assert find_version({'filters': {'__version__': None}}) is None
        assert find_version({'filters': {'__version__': [None]}}) is None
        assert find_version({'filters': {'__version__': [None, None]}}) is None
        # only the first value is used if there is a list, even if the first value is
        # invalid
        assert find_version({'filters': {'__version__': [None, 10]}}) is None

    def test_version_as_filter_exists(self):
        assert find_version({'filters': {'__version__': 10}}) == 10
        assert find_version({'filters': {'__version__': '10'}}) == 10
        assert find_version({'filters': {'__version__': [10]}}) == 10
        assert find_version({'filters': {'__version__': ['10']}}) == 10
        # only the first value is used
        assert find_version({'filters': {'__version__': ['10', None]}}) == 10

        with pytest.raises(ValueError):
            find_version({'filters': {'__version__': 'aaaaa'}})

        with pytest.raises(ValueError):
            find_version({'filters': {'__version__': ['aaaaa']}})

        with pytest.raises(ValueError):
            find_version({'filters': {'__version__': ['aaaaa', None]}})

    def test_both_are_none(self):
        assert find_version({'version': None, 'filters': {'__version__': None}}) is None
        assert (
            find_version({'version': None, 'filters': {'__version__': [None]}}) is None
        )

    def test_version_takes_precedence(self):
        assert find_version({'version': 10, 'filters': {'__version__': 12}}) == 10
        assert find_version({'version': 10, 'filters': {'__version__': [12]}}) == 10

    def test_both_mix(self):
        assert find_version({'version': 10, 'filters': {'__version__': None}}) == 10
        assert find_version({'version': 10, 'filters': {'__version__': [None]}}) == 10
        assert find_version({'version': None, 'filters': {'__version__': 12}}) == 12
        assert find_version({'version': None, 'filters': {'__version__': [12]}}) == 12


class TestMakeRequest:
    def test_empty(self):
        request = make_request({'resource_id': 'test'})
        assert request.data_dict == {'resource_id': 'test'}
        assert request.query == BasicQuery('test', None, None, None)
        assert request.size is None
        assert request.offset is None
        assert request.after is None
        assert request.sorts == []
        assert request.fields == []
        assert request.aggs == {}

    def test_with_query(self):
        request = make_request({'resource_id': 'test', 'q': 'aves'})
        assert request.data_dict == {'resource_id': 'test', 'q': 'aves'}
        assert request.query == BasicQuery('test', None, 'aves', None)

    def test_with_filters(self):
        dd = {'resource_id': 'test', 'filters': {'class': 'aves'}}
        request = make_request(dd)
        assert request.data_dict == dd
        assert request.query == BasicQuery('test', None, None, dd['filters'])

    def test_with_options(self):
        dd = {
            'resource_id': 'test',
            'limit': 4,
            'offset': 9,
            'after': 'someaftervalue',
            'sort': ['field1 asc', 'field2 desc'],
            'fields': ['field4', 'field5'],
        }
        request = make_request(dd)
        assert request.data_dict == dd
        assert request.query == BasicQuery('test', None, None, None)
        assert request.size == 4
        assert request.offset == 9
        assert request.after == 'someaftervalue'
        assert request.sorts == [Sort('field1'), Sort('field2', False)]
        assert request.fields == ['field4', 'field5']

    # todo: add facet test



from ckan.plugins import toolkit
from jsonschema.exceptions import ValidationError

from ckanext.versioned_datastore.lib.query.search.query import SchemaQuery
from ckanext.versioned_datastore.lib.query.search.request import SearchRequest
from ckanext.versioned_datastore.lib.utils import get_available_datastore_resources


def make_request(data_dict: dict) -> SearchRequest:
    """
    Given an action data dict, creates a SearchRequest object for it.

    :param data_dict: the action data dict
    :returns: a SearchRequest object
    """
    # if no resource IDs are provided, default to all resources available to the user
    resource_ids = data_dict.get('resource_ids', [])
    if not resource_ids:
        resource_ids = list(get_available_datastore_resources())

    query = SchemaQuery(
        resource_ids,
        data_dict.get('version'),
        data_dict.get('query'),
        data_dict.get('query_version'),
    )

    # check the query is valid before we go any further
    try:
        query.validate()
    except ValidationError as e:
        raise toolkit.ValidationError(e.message)

    request = SearchRequest(
        query,
        size=data_dict.get('size'),
        after=data_dict.get('after'),
        data_dict=data_dict,
    )

    # ignore any resources that are unavailable for whatever reason
    request.add_param('ignore_unavailable', True)

    return request



from .csv import CsvDerivativeGenerator
from .dwc import DwcDerivativeGenerator
from .json import JsonDerivativeGenerator

derivatives = [
    CsvDerivativeGenerator,
    JsonDerivativeGenerator,
    DwcDerivativeGenerator,
]



from ckanext.datastore.logic.schema import json_validator
from ckanext.versioned_datastore.logic.validators import (
    ignore_missing,
    int_validator,
    not_empty,
    not_missing,
    resource_id_exists,
)


def vds_options_get() -> dict:
    return {
        'resource_id': [not_empty, str, resource_id_exists],
        'version': [ignore_missing, int_validator],
    }


def vds_options_update() -> dict:
    return {
        'resource_id': [not_empty, str, resource_id_exists],
        'overrides': [not_missing, json_validator],
    }



import json
from copy import deepcopy
from typing import Dict

from ckan.plugins import toolkit
from splitgill.utils import now

from ckanext.versioned_datastore.lib import common
from ckanext.versioned_datastore.lib.query.search.query import SchemaQuery
from ckanext.versioned_datastore.lib.utils import (
    get_available_resources,
    get_database,
    ivds_implementations,
)


def get_resources_and_versions(
    query: SchemaQuery,
    allow_non_datastore: bool = False,
) -> Dict[str, int]:
    """
    Given a query on some resources, returns a dict of resource IDs to versions which
    should be queried based on the overall query version. The version provided with each
    resource ID in the dict will be the rounded version, representing the version of the
    data which should be queried to get the data at the overall version.

    This function also filters out resources the user doesn't have access to.

    :param query: a SchemaQuery object representing the query
    :param allow_non_datastore: allow non datastore resources to be included (will be
        returned with common.NON_DATASTORE_VERSION)
    :returns: a dict of resource IDs and versions
    """
    available_resource_ids = get_available_resources(datastore_only=False)
    if not available_resource_ids.issuperset(query.resource_ids):
        raise toolkit.ValidationError(
            'Not all requested resources are accessible to this user'
        )

    version = now() if query.version is None else query.version
    resource_ids_and_versions = {}
    for resource_id in query.resource_ids:
        # round the version down to ensure we search the exact version requested
        rounded_version = get_database(resource_id).get_rounded_version(version)
        # resource ids without a rounded version are skipped
        if rounded_version is not None:
            resource_ids_and_versions[resource_id] = rounded_version
        elif allow_non_datastore:
            # technically, assuming that a resource with no version available is a
            # non-datastore resource can be wrong because it could just mean that the
            # resource wasn't created at the time of the overall version and the request
            # is before the resource existed in the system at all. However, in practice
            # this isn't worth the hassle of dealing with because it's really unlikely
            # to occur and if it does something funny is happening elsewhere.
            resource_ids_and_versions[resource_id] = common.NON_DATASTORE_VERSION

    return resource_ids_and_versions


def convert_to_multisearch(query: dict) -> dict:
    """
    Converts the given basic query dict into a multisearch query and returns it.

    :param query: a basic query dict
    :returns: a multisearch query dict
    """
    # save a copy of the original query
    basic_query = deepcopy(query)
    multisearch_query = {}

    # allow other plugins to modify the query before processing, e.g. to remove any
    # custom filters
    for plugin in ivds_implementations():
        query = plugin.datastore_before_convert_basic_query(query)

    if 'q' in query:
        multisearch_query['search'] = query['q']

    if 'filters' in query:
        filter_list = []
        for field, values in query['filters'].items():
            if not isinstance(values, list):
                values = [values]
            if field == '__geo__':
                for value in values:
                    if isinstance(value, str):
                        value = json.loads(value)
                    if value['type'] == 'Polygon':
                        filter_list.append({'geo_custom_area': [value['coordinates']]})
                    else:
                        # I cannot find any examples of anything other than polygons, so
                        # I'm not sure it was ever implemented for the old searches
                        raise NotImplemented
            else:
                subgroup = []
                subgroup_count = 0
                for value in values:
                    if field != '' and value != '':
                        subgroup.append(
                            {'string_equals': {'fields': [field], 'value': value}}
                        )
                        subgroup_count += 1
                if subgroup_count > 1:
                    filter_list.append({'or': subgroup})
                elif subgroup_count == 1:
                    filter_list += subgroup
        multisearch_query['filters'] = {'and': filter_list}

    # allow plugins to modify the output, with the additional context of the original
    # basic query
    for plugin in ivds_implementations():
        multisearch_query = plugin.datastore_after_convert_basic_query(
            basic_query, multisearch_query
        )

    return multisearch_query


def convert_small_or_groups(query):
    """
    Convert OR groups containing only 1 item to AND groups.

    :param query: a multisearch query dict
    :returns: the query with a converted filter dict, if applicable
    """
    if 'filters' not in query:
        return query

    def _convert(*filters):
        items = []
        for term_or_group in filters:
            k, v = list(term_or_group.items())[0]
            if k not in ['and', 'or', 'not']:
                items.append(term_or_group)
            elif k != 'or' or len(v) != 1:
                # don't convert empty groups because those throw an error for all types
                items.append({k: _convert(*v)})
            else:
                items.append({'and': _convert(*v)})
        return items

    query['filters'] = _convert(query['filters'])[0]

    return query


def remove_empty_groups(query):
    """
    Remove empty groups from filter list.

    :param query: a multisearch query dict
    :returns: the query with a processed filter dict, if applicable
    """
    if 'filters' not in query:
        return query

    def _convert(*filters):
        items = []
        for term_or_group in filters:
            k, v = list(term_or_group.items())[0]
            if k not in ['and', 'or', 'not']:
                items.append(term_or_group)
            elif len(v) > 0:
                items.append({k: _convert(*v)})
        return items

    processed_filters = _convert(query['filters'])
    if len(processed_filters) == 0:
        del query['filters']
    else:
        query['filters'] = processed_filters[0]

    return query



import abc
import logging
import tempfile
from pathlib import Path
from typing import List, Optional, Union

from cachetools import TTLCache, cached
from ckan.plugins import toolkit
from rq.job import Job

from ckanext.versioned_datastore.lib.utils import es_client


class Task(abc.ABC):
    """
    Abstract base class for VDS queued tasks, for example importing downloading data.
    """

    def __init__(
        self,
        queue_name: str,
        title: str,
        timeout: int = 3600,
    ):
        """
        :param queue_name: the name of the queue this task should run on
        :param title: the title of the task
        :param timeout: how long RQ should wait for this task to complete, in seconds,
                        before timing it out (defaults to 3600 which is 1 hour)
        """
        self.queue_name = queue_name
        self.title = title
        self.timeout = timeout
        self.log = logging.getLogger(__name__)

    @abc.abstractmethod
    def run(self, tmpdir: Path):
        """
        Method stub for subclasses to override. This is automatically called by start
        when a task begins after some setup is completed.

        :param tmpdir: a temporary directory which can be used by the task and will be
            cleaned up after this run method completes
        """
        ...

    def start(self):
        """
        Starts the task, performs some setup and then calls self.run with a temporary
        directory.
        """
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            self.log.info(f'Starting {self.title}')
            self.run(Path(tmp_dir_name))
        self.log.info(f'Finished {self.title}')

    def queue(
        self,
        depends_on: Optional[Union[Job, List[Job]]] = None,
        timeout: Optional[int] = None,
    ) -> Job:
        """
        Queues this task on the appropriate RQ queue.

        :param depends_on: a job, or list of jobs, which this task depends on. If given,
            this task will only start when the other jobs finish successfully. Check the
            RQ docs for more information.
        :param timeout: a timeout in seconds for this task. If provided this overrides
            the base timeout defined in the task, otherwise the default task's timeout
            is used.
        :returns: returns the result of calling CKAN's enqueue_job which will provide
            details about the queued job for this task
        """
        rq_kwargs = {'timeout': timeout or self.timeout}
        if depends_on:
            rq_kwargs['depends_on'] = depends_on

        return toolkit.enqueue_job(
            self.start,
            queue=self.queue_name,
            title=self.title,
            rq_kwargs=rq_kwargs,
        )


@cached(cache=TTLCache(maxsize=10, ttl=300))
def get_queue_length(queue_name):
    """
    This will only get the pending jobs in a queue, not any jobs that are currently
    processing.

    :param queue_name: the name of the queue to check, e.g. 'download'
    :returns: length of queue as int
    """
    queued_jobs = toolkit.get_action('job_list')(
        {'ignore_auth': True}, {'queues': [queue_name]}
    )
    return len(queued_jobs)


@cached(cache=TTLCache(maxsize=10, ttl=300))
def get_es_health():
    client = es_client()
    return {'ping': client.ping(), 'info': client.info()}



from typing import Optional

from ckantools.decorators import action

from ckanext.versioned_datastore.lib.importing.options import (
    get_options,
    update_options,
)
from ckanext.versioned_datastore.lib.utils import (
    ReadOnlyResourceException,
    is_resource_read_only,
)
from ckanext.versioned_datastore.logic.options import helptext, schema


@action(schema.vds_options_get(), helptext.vds_options_get, get=True)
def vds_options_get(resource_id: str, version: Optional[int] = None) -> Optional[dict]:
    """
    Retrieves the options in use by Splitgill for the given resource at the given
    optional version.

    :param resource_id: the resource ID
    :param version: the version (default is None which means get the latest options)
    :returns: the options as a dict, or None if there are no options available
    """
    options = get_options(resource_id, version)
    return None if options is None else options.to_doc()


@action(schema.vds_options_update(), helptext.vds_options_update)
def vds_options_update(
    resource_id: str,
    overrides: dict,
) -> Optional[int]:
    """
    Updates the options for the given resource with the new values in the overrides
    parameter. The existing options are retrieved and then the new options in the
    overrides dict are written over the top. This allows incremental updates to the
    existing options. If no options exist, the default options are used as a base.

    :param resource_id: the resource ID
    :param overrides: a dict of override options to apply
    :returns: the new options version, or None if nothing was changed
    """
    if is_resource_read_only(resource_id):
        raise ReadOnlyResourceException('This resource has been marked as read only')

    return update_options(resource_id, overrides)



from ckanext.versioned_datastore.logic.validators import (
    boolean_validator,
    ignore_missing,
    int_validator,
    not_empty,
    not_missing,
    resource_id_exists,
    validate_datastore_resource_id,
)


def vds_data_add() -> dict:
    return {
        'resource_id': [not_empty, str, resource_id_exists],
        'replace': [not_missing, boolean_validator],
    }


def vds_data_delete() -> dict:
    return {
        'resource_id': [not_empty, str, resource_id_exists],
    }


def vds_data_sync() -> dict:
    return {
        'resource_id': [not_empty, str, resource_id_exists],
        'full': [ignore_missing, boolean_validator],
    }


def vds_data_get() -> dict:
    return {
        'resource_id': [not_empty, str, validate_datastore_resource_id],
        'record_id': [not_empty, str],
        'version': [ignore_missing, int_validator],
    }



from operator import itemgetter

import pytest
from splitgill.model import Record

from ckanext.versioned_datastore.lib.downloads import utils
from ckanext.versioned_datastore.lib.downloads.utils import get_schema
from ckanext.versioned_datastore.lib.importing.options import (
    create_default_options_builder,
)
from ckanext.versioned_datastore.lib.utils import get_database


class TestGetSchema:
    @pytest.mark.usefixtures('with_vds')
    def test_get_schema(self):
        resource_id = 'test-resource-id'
        database = get_database(resource_id)
        database.update_options(create_default_options_builder().build(), commit=False)
        database.ingest(
            [
                Record.new(
                    {
                        'name': 'Paru',
                        'size': 5,
                        'dob': '2021-01-12',
                        'good': True,
                        'toys': ['Feather stick', 'Monsieur Canard', 'Catnip Carrot'],
                        'food': [
                            {
                                'name': 'Chicken',
                                'weight': 40,
                            },
                            {
                                'name': 'Kibble',
                                'weight': 36,
                            },
                        ],
                    }
                )
            ],
            commit=True,
        )
        database.sync()

        schema = get_schema(resource_id)

        assert isinstance(schema, dict)
        assert schema['type'] == 'record'
        assert schema['name'] == 'Record'
        assert len(schema['fields']) == 7

        assert sorted(schema['fields'], key=itemgetter('name')) == [
            {'name': '_id', 'type': ['string', 'null']},
            {'name': 'dob', 'type': ['string', 'null']},
            {
                'name': 'food',
                'type': [
                    {
                        'items': [
                            [
                                {
                                    'fields': [
                                        {'name': 'name', 'type': ['string', 'null']},
                                        {'name': 'weight', 'type': ['long', 'null']},
                                    ],
                                    'name': 'food.Record',
                                    'type': 'record',
                                },
                                'null',
                            ]
                        ],
                        'type': 'array',
                    },
                    'null',
                ],
            },
            {'name': 'good', 'type': ['boolean', 'null']},
            {'name': 'name', 'type': ['string', 'null']},
            {'name': 'size', 'type': ['long', 'null']},
            {
                'name': 'toys',
                'type': [{'items': [['string', 'null']], 'type': 'array'}, 'null'],
            },
        ]


class TestFilterDataFields:
    def test_excludes_field_with_zero_count(self):
        data = {'one': 'a', 'two': 'b', 'three': 'c'}
        field_counts = {'one': 1, 'two': 1, 'three': 0}

        filtered_data = utils.filter_data_fields(data, field_counts)
        assert 'one' in filtered_data
        assert 'two' in filtered_data
        assert 'three' not in filtered_data

    def test_excludes_field_with_no_count(self):
        data = {'one': 'a', 'two': 'b', 'three': 'c'}
        field_counts = {'one': 1, 'two': 1}

        filtered_data = utils.filter_data_fields(data, field_counts)
        assert 'one' in filtered_data
        assert 'two' in filtered_data
        assert 'three' not in filtered_data

    def test_includes_null_field_with_count(self):
        data = {'one': 'a', 'two': 'b', 'three': None}
        field_counts = {'one': 1, 'two': 1, 'three': 1}

        filtered_data = utils.filter_data_fields(data, field_counts)
        assert 'one' in filtered_data
        assert 'two' in filtered_data
        assert 'three' in filtered_data

    def test_excludes_empty_nested_fields(self):
        data = {'one': {'two': 'a', 'three': 'b'}}
        field_counts = {'one.two': 1}

        filtered_data = utils.filter_data_fields(data, field_counts)
        assert 'one' in filtered_data
        assert 'two' in filtered_data['one']
        assert 'three' not in filtered_data['one']

    def test_excludes_empty_listed_fields(self):
        data = {'one': [{'two': 'a', 'three': 'b'}, {'two': 'c', 'four': 'd'}]}
        field_counts = {'one.two': 2, 'one.four': 1}

        filtered_data = utils.filter_data_fields(data, field_counts)
        assert 'one' in filtered_data
        assert len(filtered_data['one']) == 2
        assert 'four' in filtered_data['one'][1]
        for r in filtered_data['one']:
            assert 'two' in r
            assert 'three' not in r


class TestFlattenDict:
    def test_flattens_dict(self):
        data = {'one': {'two': 2, 'three': 3}}
        expected_output = {'one.two': 2, 'one.three': 3}

        flattened = utils.flatten_dict(data)
        assert len(flattened) == len(expected_output)
        for k, v in expected_output.items():
            assert k in flattened
            assert v == flattened[k]

    def test_flattens_nested_lists(self):
        data = {'one': [{'two': 'a', 'three': 'b'}, {'two': 'c', 'four': 'd'}]}
        expected_output = {'one.two': 'a | c', 'one.three': 'b', 'one.four': 'd'}

        flattened = utils.flatten_dict(data)
        assert len(flattened) == len(expected_output)
        for k, v in expected_output.items():
            assert k in flattened
            assert v == flattened[k]

    def test_flattens_lists(self):
        data = {'one': ['a', 'b']}
        expected_output = {'one': 'a | b'}

        flattened = utils.flatten_dict(data)
        assert len(flattened) == len(expected_output)
        for k, v in expected_output.items():
            assert k in flattened
            assert v == flattened[k]



from unittest.mock import MagicMock, patch

import pytest
from ckan.plugins import toolkit

from ckanext.versioned_datastore.logic.validators import (
    _deduplicate,
    validate_datastore_resource_ids,
    validate_resource_ids,
)


class TestDeduplicate:
    def test_empty(self):
        assert not list(_deduplicate([]))

    def test_all_different(self):
        assert list(_deduplicate(['a', 'b', 'c'])) == ['a', 'b', 'c']

    def test_some_duplicates_and_order_preserved(self):
        assert list(_deduplicate(['a', 'd', 'a', 'b', 'a', 'c', 'b'])) == [
            'a',
            'd',
            'b',
            'c',
        ]


class TestValidateResourceIDs:
    def test_empty(self):
        assert validate_resource_ids([]) == []
        assert validate_resource_ids('') == []

    def test_not_a_list(self):
        with pytest.raises(
            toolkit.Invalid, match='Invalid list of resource ID strings'
        ):
            validate_resource_ids({})

    def test_all_valid(self):
        valid = ['a', 'c', 'x']
        check_resource_id_mock = MagicMock(side_effect=lambda r, _: r in valid)

        with patch(
            'ckanext.versioned_datastore.logic.validators.check_resource_id',
            check_resource_id_mock,
        ):
            assert validate_resource_ids(valid) == valid

    def test_all_invalid(self):
        check_resource_id_mock = MagicMock(return_value=False)

        with patch(
            'ckanext.versioned_datastore.logic.validators.check_resource_id',
            check_resource_id_mock,
        ):
            with pytest.raises(toolkit.Invalid, match='No resource IDs are available'):
                validate_resource_ids(['o', 'b'])

    def test_some_valid(self):
        valid = {'a', 'c', 'x'}
        check_resource_id_mock = MagicMock(side_effect=lambda r, _: r in valid)

        with patch(
            'ckanext.versioned_datastore.logic.validators.check_resource_id',
            check_resource_id_mock,
        ):
            assert validate_resource_ids(['x', 'o', 'a', 'b', 'c']) == ['x', 'a', 'c']

    def test_some_valid_with_duplicates(self):
        valid = {'a', 'c', 'x'}
        check_resource_id_mock = MagicMock(side_effect=lambda r, _: r in valid)

        with patch(
            'ckanext.versioned_datastore.logic.validators.check_resource_id',
            check_resource_id_mock,
        ):
            assert validate_resource_ids(['x', 'a', 'a', 'a', 'b']) == ['x', 'a']


class TestValidateDatastoreResourceIDs:
    def test_empty(self):
        assert validate_datastore_resource_ids([]) == []
        assert validate_datastore_resource_ids('') == []

    def test_not_a_list(self):
        with pytest.raises(
            toolkit.Invalid, match='Invalid list of resource ID strings'
        ):
            validate_datastore_resource_ids({})

    def test_all_valid(self):
        valid = ['a', 'c', 'x']
        check_datastore_resource_id_mock = MagicMock(
            side_effect=lambda r, _: r in valid
        )

        with patch(
            'ckanext.versioned_datastore.logic.validators.check_datastore_resource_id',
            check_datastore_resource_id_mock,
        ):
            assert validate_datastore_resource_ids(valid) == valid

    def test_all_invalid(self):
        check_datastore_resource_id_mock = MagicMock(return_value=False)

        with patch(
            'ckanext.versioned_datastore.logic.validators.check_datastore_resource_id',
            check_datastore_resource_id_mock,
        ):
            with pytest.raises(
                toolkit.Invalid, match='No resource IDs are datastore resources'
            ):
                validate_datastore_resource_ids(['o', 'b'])

    def test_some_valid(self):
        valid = {'a', 'c', 'x'}
        check_datastore_resource_id_mock = MagicMock(
            side_effect=lambda r, _: r in valid
        )

        with patch(
            'ckanext.versioned_datastore.logic.validators.check_datastore_resource_id',
            check_datastore_resource_id_mock,
        ):
            assert validate_datastore_resource_ids(['x', 'o', 'a', 'b', 'c']) == [
                'x',
                'a',
                'c',
            ]

    def test_some_valid_with_duplicates(self):
        valid = {'a', 'c', 'x'}
        check_datastore_resource_id_mock = MagicMock(
            side_effect=lambda r, _: r in valid
        )

        with patch(
            'ckanext.versioned_datastore.logic.validators.check_datastore_resource_id',
            check_datastore_resource_id_mock,
        ):
            assert validate_datastore_resource_ids(['x', 'a', 'a', 'a', 'b']) == [
                'x',
                'a',
            ]



from unittest.mock import patch

import pytest

from ckanext.versioned_datastore.logic.download.action import vds_download_queue
from ckanext.versioned_datastore.logic.download.arg_objects import (
    DerivativeArgs,
    NotifierArgs,
    QueryArgs,
)
from tests.helpers import patches


class TestQueueDownload:
    @pytest.mark.usefixtures('with_vds')
    def test_queue_direct_call(self):
        # there is a very similar test in test_downloads.py that calls this via the API
        # instead
        resource_ids = ['test-resource-id']

        with patch('ckan.plugins.toolkit.enqueue_job') as enqueue_mock:
            with patches.rounded_versions():
                with patches.get_available_resources(resource_ids):
                    with patches.url_for():
                        vds_download_queue(
                            {},
                            QueryArgs(),
                            DerivativeArgs(format='csv'),
                            notifier=NotifierArgs(type='none'),
                        )
                        assert enqueue_mock.call_count == 1
